/******************************************************************************
* Filename: dice.c
* Description: Main source of DiCe
* Author: [Jedi] (jedi_il@hotmail.com)
* Copyright 2002 [Jedi]
* See COPYING for the GPL GNU
*
*    This file is part of [Jedi]'s DiCe.
*
*    [Jedi]'s DiCe is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    [Jedi]'s DiCe is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with Foobar; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/

#include "diceconf.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>

/**** Common outputs ***************/
#define GAME_NOT_STARTED "Game not started, ask one of the siteops to start the game.\n"

/**** Constants ********************/
#define VERSION "1.02"
#define TAG "-- [Jedi] DiCe Script v1.02 --\n"
#define LEN 1024
#ifndef MAXPATHLEN
#define MAXPATHLEN 512
#endif

/**** Linked List Structure ********/

struct gambler {
	char name[24];
	char group[24];
	int g1;
	int g2;
	int winner;
	unsigned long credits; /* in Kb (from config file */
	int g_creds; /* in Mb */
	struct gambler *next;
};
typedef struct gambler gambler;

/******* Linked List Functions *****/
gambler *new_user();
int load_users();
void free_list();

/******* Global Variables **********/
int dn; 	/* Dice Number */
struct gambler *anchor=NULL;     /* Anchor */
int total_gamblers=0, 
    winners=0,
    jackpot=0,
	win_creds=0;

char line[LEN], path[MAXPATHLEN], username[24], flags[24], group[24];

/******* Functions *****************/
unsigned long get_credits(const char *username);
void strsplit_s(char *output, const char *str, const char separator, int index);
int is_started();
int ratio = -1;
void rolldice(FILE *stream);
void printhelp(int argc, char **argv);
void addbet(int argc, char *args[], int index);
void change_credits(char *username, float change_mb);
void startgame(FILE *stream);
void killgame();
void undice();
void showbetlist();
void gl_log(char *logline);

/*******************************************
* Fucntion: main()
*******************************************/

int main(int argc, char **argv) {

	strcpy(username,getenv("USER"));
	if (getenv("FLAGS")!=NULL) strcpy(flags,getenv("FLAGS"));

	if (getenv("GROUP")!=NULL) strcpy(group,getenv("GROUP"));
	else strcpy(group,"NoGroup");

	if (getenv("RATIO")!=NULL) ratio = atoi(getenv("RATIO"));
	else ratio = -1;

	if (argc==1) { 
		printhelp(argc, argv);
		exit(0); 
	}
	
	if (strcasecmp(argv[1],"roll") == 0 && strstr(flags,MASTER_FLAG)!=NULL) rolldice(stdout);
	else if (strcasecmp(argv[1],"start") == 0 && strstr(flags,MASTER_FLAG)!=NULL) startgame(stdout);
	else if (strcasecmp(argv[1],"kill") == 0 && strstr(flags,MASTER_FLAG)!=NULL) killgame();
	else if (strcasecmp(argv[1],"list") == 0) showbetlist();
	else if (strcasecmp(argv[1],"unbet") == 0) undice();
	else if (strcasecmp(argv[1],"bet") == 0) addbet(argc,argv,2);
	else printhelp(argc, argv);

	return 0;
}


/************************************************
* Function: startgame()
* Desc    : Initiate a new game by writing zero 
*	        length file in DICEDB
************************************************/

void startgame(FILE *stream) {
	FILE *newgame;

	if (is_started()) {
		fprintf(stream, TAG);
		fprintf(stream, "A game is already started, type \"site dice roll\" to play (roll dices), and end the game.\n");
		return;
	}
	newgame=fopen(DICEDB,"w");
	if (newgame==NULL) {
		fprintf(stream, "Error %d: %s (%s)\n",errno,strerror(errno),DICEDB);
		return;
	}
	fclose(newgame);
	chmod(DICEDB,0666);
	fprintf(stream, TAG);
	fprintf(stream, "A new game commenced successfully!\n");
	fprintf(stream, "Wait for users to bet, use \"site dice list\" to show the bets taken so far\n");
	fprintf(stream, "Use \"site dice roll\" to play - roll the dices, and end the game\n");

	sprintf(line,"NEWDICE: \"%s\" \"%s\" \"%d\"",username, group, DICESIZE);
	gl_log(line);
}


/************************************************
* Function: killgame()
* Desc    : Kills the game (without rolling dice)
************************************************/

void killgame() {
	printf(TAG);		
	if (!is_started()) {
		printf("Game not started, there's nothing to kill.\n");
		return;
	}
	if (unlink(DICEDB) == -1)
		printf("Error %d: %s (%s)\n",errno,strerror(errno),DICEDB);
	else
		printf("Game killed successfully.\n");

}

/*********************************************
* Function: addbet()
* Gets    : argc, argv, index
* Desc    : add's a user bet to the DICEDB
*********************************************/

void addbet(int argc, char *args[], int index) {
	int gamble_mb=0, ga1=0, ga2=0, k;
	float b;
	FILE *dicedb;
	#ifdef LIMIT_FLAGS
		char limit_flags[] = LIMIT_FLAGS;
	#endif
	#if (AUTOROLL_MB > 0 || AUTOROLL_USERS > 0)
		FILE *stream;
	#endif

	printf(TAG);
	if (!is_started()) {
		printf(GAME_NOT_STARTED);
		return;
	}
	if (argc < 4) {
		printf("Missing arguments, usage \"site dice bet <credits> <gamble1> [gamble2]\"\n");
		printf("Type \"site dice\" for more info\n");
		return;
	}
	gamble_mb = atoi(args[index]);
	ga1 = atoi(args[index+1]);

	#ifdef LIMIT_LEECH
	  if (ratio == 0) {
		if (LIMIT_MB == 0) {
			printf("Leech users cannot bet!\n");
			return;
		}
		else if (LIMIT_MB < gamble_mb) {
			printf("Cannot add your bet!\n");
			printf("Leech users may gamble up to %dMb per bet.\n",LIMIT_MB);
			return;
		}
	  }
	#endif

	#ifdef LIMIT_FLAGS
	  for (k=0; k < sizeof(limit_flags)-1; k++) {
		if (strchr(flags,limit_flags[k])) {
			if (LIMIT_MB == 0) {
				printf("Users with %c flag cannot bet!\n", limit_flags[k]);
				return;
			}
			else if (LIMIT_MB < gamble_mb) { 
				printf("Cannot add your bet!\n");
				printf("Users with %c flag may gamble up to %dMb per bet.\n",limit_flags[k],LIMIT_MB);
				return;
			}
		}
	  }
	#endif

	if (argc > (index+2)) {
		ga2 = atoi(args[index+2]);
		if (ga2 > DICESIZE || ga2 < 1) {
			printf("Error, dice 2 is out of range, choose a number between 1 to %d !\n", DICESIZE);
			exit(0);
		}
	} 
	if (ga1 > DICESIZE || ga1 < 1) { 
		printf("Error, dice is out of range, choose a number between 1 to %d !\n", DICESIZE);
		exit(0);
	}
	if (gamble_mb < MIN_GAMBLE || gamble_mb > MAX_GAMBLE) {
		printf("Your gabmle is out of range, it must be between %d to %d (megabytes)\n", MIN_GAMBLE, MAX_GAMBLE);
		exit(0);
	}

	/* We double the gamble_mb now, becuase we want it to pass the min_gamble and max_gample
	   limitations, but not pass the user's credits amount limitation, and we want it to be
	   written to the database plainly */

	if (ga2!=0) gamble_mb*=2;

	b=(float)get_credits(username)/1024.0;

	if (b < (float)gamble_mb) {
		printf("You can't gamble on more than %.2fMb credits which you have!\n",b);
		return;
	}

	dicedb=fopen(DICEDB,"r+"); /* open for reading/writing */
	if (dicedb==NULL) {
		printf("Error %d: %s (%s)\n",errno,strerror(errno),DICEDB);
		return;
	}
	k=strlen(username);
	while (fgets(line,LEN-1,dicedb)!=NULL) {
		if (strncmp(username,line,k)==0) {
			printf("You've already betted this round, if you want to remove your bet use \"site dice unbet\"\n");
			fclose(dicedb);
			return;
		}
	}
	fprintf(dicedb,"%s %s %d %d %d\n",username,group,gamble_mb,ga1,ga2);
	fclose(dicedb);
	printf("Your bet has been added successfully!\n");
	printf("You've betted on (%dMb) credits\n on the number(s): (%d)",gamble_mb,ga1);
	if (ga2!=0) printf(" and (%d)",ga2);
	printf("\nIf you wish to undo your bet, you can simply \"site dice unbet\"\n");
	printf("\nGood luck!\n");

	if (ga2==0) sprintf(line,"DICE: \"%s\" \"%s\" \"%d\" \"%d\"",username,group,gamble_mb,ga1);
	else sprintf(line,"DICE: \"%s\" \"%s\" \"%d\" \"%d & %d\"",username,group,gamble_mb,ga1,ga2);
	gl_log(line);
	#if (AUTOROLL_MB > 0 || AUTOROLL_USERS > 0)
		if (load_users()) return;
		if ( (AUTOROLL_MB > 0    && jackpot >= AUTOROLL_MB) || 
		     (AUTOROLL_USERS > 0 && total_gamblers >= AUTOROLL_USERS ) ) {
			free_list();
			strcpy(username,A_USER);
			strcpy(group,A_GROUP);
			if (strstr(flags,MASTER_FLAG) != NULL)	rolldice(stdout);
			else {
				stream = fopen("/dev/null","w");
				if (stream != NULL) {
					rolldice(stream);
					fclose(stream);
				}
			}
		} else free_list();
	#endif
}


/********************************************
* Function: undice()
* Desc    : removes a user's bet from DICEDB
********************************************/

void undice() {
	FILE *dicedb, *tmpfile;
	int fd, k, found=0;

	printf(TAG);
	if (!is_started()) {
		printf(GAME_NOT_STARTED);
		return;
	}
	k=strlen(username);

	sprintf(path,"%s.XXXXXX",DICEDB);
	fd = mkstemp(path);    /* Create temp file to write to */
	if( fd == -1 ) {                  
		printf("Cant open '%s' for writing!\n",path);
		return;
	}
	tmpfile = fdopen(fd,"w");
	if (tmpfile==NULL) {
		printf("Error %d: %s (%s tempfile)\n",errno,strerror(errno),DICEDB);
		return;
	}

	dicedb=fopen(DICEDB,"r");
	if (dicedb==NULL) {
		printf("Error %d: %s (%s)\n",errno,strerror(errno),DICEDB);
		return;
	}
	while (fgets(line,LEN-1,dicedb)!=NULL) {
		if (strncmp(username,line,k)==0) found=1;
		else fputs(line,tmpfile);
	}
	fclose(dicedb);
	fclose(tmpfile);

	rename(path,DICEDB); /* move tmpfile over dicedb */
	chmod(DICEDB,0666);
	if (found==1) {
		printf("You'r bet was successfully removed!\n");
		sprintf(line,"UNDICE: \"%s\" \"%s\"",username,group);
		gl_log(line);
	}
	else printf("Could not remove your bet, make sure you've put one at all...\n");
	
}


/**********************************************
* Function: showbetlist()
* Desc    : Outputs to STDOUT the bookie i.e
*           the betting list...
**********************************************/

void showbetlist() {
	gambler *user;	

	if (!is_started()) {
		printf(TAG);
		printf(GAME_NOT_STARTED);
		return;
	}
	if (load_users()) return;
	printf(" ,-----------+++`````( The List )'''''+++-------------, \n");
	for(user=anchor; user!=NULL; user=user->next) {
		if (user->g2==0)
			printf("| %10s gambles on ( %2d  ) for %7dMb credits |\n", 
			        user->name, user->g1, user->g_creds);
		else
			printf("| %10s gambles on (%-2d %2d) for %7dMb credits |\n", user->name, user->g1, 
			       user->g2, user->g_creds);
	}
	printf(" `----------------------------------------------------'\n");
	printf("   Current Jackpot: (%dMb)\n", jackpot);
	printf("   Dice Size      : (%d)\n", DICESIZE);
	printf(TAG);
	free_list();

}


/***********************************************
* Function: printhelp()
* Desc    : Outputs the help screen
***********************************************/
 	
void printhelp(int argc, char **argv) {
	printf("**** [Jedi]'s DiCe Script v%s ****\n",VERSION);
	printf("Usage: site dice <command> [args...]\n\
Dice Size: %d (bets must be between 1 to %d)\n\
site dice bet <credits> <bet1> [bet2]  - puts a gamble on bet1 and optionally bet2 for credits Mb,\
   if you put 2 bets, it will be as if you'be bet on 2xCredits Mb, but it also increase your chances!\n\
For example: \"site dice bet 500 4\" bets on 4 for 500mb\n\
             \"site dice bet 500 4 6\" bets on 4 and 6 for 1000mb\n\
site dice unbet  - removes your bet (only if you have betted this round)\n\
site dice list   - shows the bookies (the betting list) of users, number and how much they bet on!\n\
site dice roll   - (*) will run the game, give and take credits from winners and loosers\n\
site dice start  - (*) will commence a dice game, users then may use \"site dice bet\" \n\
site dice kill   - (*) will kill a started game (abort the game), if one started\n\
(*) - only users with the flag %s may run the command\n\
(visit my website @ http://jedi.says.no )\n", DICESIZE, DICESIZE, MASTER_FLAG);

}


/*********************************************************
* Function: rolldice()
* Desc    : rolls two random dice, and handles the winners
*           and the losers
*********************************************************/

void rolldice(FILE *stream) {
	gambler *user;
	float winner_gets, winner_jackpot;
	struct stat st;
	FILE *dicejp;

	srand(getpid());        /* Randomize (generate random seed); */
	dn = 1+(int)((float)DICESIZE * rand()/(RAND_MAX+1.0)); /* Set Dice Number  */

	if (load_users()) {
		fprintf(stream, TAG);
		fprintf(stream, "Could not roll: Make sure a new game is started !\n");
		return;
	}
	if (total_gamblers < MIN_GAMBLERS) {
		fprintf(stream, "Only %d users have places bets, while %d are needed, have more users place bets\n",
		        total_gamblers, MIN_GAMBLERS);
		free_list();
		return;
	}
 

	fprintf(stream, "Dice (1-%d) result...\n",DICESIZE);
	fprintf(stream, "Dice Winning Number : -=(%d)=-\n",dn);

	sprintf(line,"ROLLDICE: \"%s\" \"%s\" \"%d\"",username,group,dn);
	gl_log(line);	

	if (winners==0) {
		fprintf(stream, "Nobody won!\n");
		dicejp=fopen(DICEJP,"r");
		line[0]='0'; line[1]='\0';
		if (dicejp!=NULL) {
			fgets(line,LEN-1,dicejp);
			fclose(dicejp);
		}
		dicejp=fopen(DICEJP,"w");
		if (dicejp!=NULL) {
			fprintf(dicejp,"%d",atoi(line)+(int)(jackpot/2));
			fclose(dicejp);
			chmod(DICEJP,0666);
		}
	} else {
		stat(DICEJP,&st);
		if (S_ISREG(st.st_mode)) /* if file exists */
			if (unlink(DICEJP) == -1) /* Clean the jackpot */
				fprintf(stream, "Error %d: Could not clean the jackpot (%s), do it manually!\n",errno,strerror(errno));
	}

	for(user=anchor; user!=NULL; user=user->next) {
		if (user->winner==1) {

			/* winner_jackpot - the amount of credits the user get from the jackpot (% of the jackpot)
			   winner_gets - the actual number that user credits is being increased by */
			winner_jackpot = (float)( (float)jackpot * (float)user->g_creds / (float)win_creds );
			winner_gets = winner_jackpot - (NO_INFLATION_FLAG * (float)user->g_creds);
			change_credits(user->name,winner_gets);

			fprintf(stream, "%s is a WINNER!\n",user->name);
			sprintf(line, "DICEWIN: \"%s\" \"%s\" \"%.1f\" \"%.1f\" \"%.1f\"", user->name, user->group,
				     winner_jackpot, (float)(100*(float)user->g_creds/(float)win_creds), winner_gets);
			gl_log(line);
		} else  {
			change_credits(user->name,(float)(-1 * user->g_creds) );
			fprintf(stream, "%s is a LOSER!\n",user->name);
		}
	}
	
	if (unlink(DICEDB) == -1)
		fprintf(stream, "Error %d: %s (%s)\n",errno,strerror(errno),DICEDB);

	if (winners==0) {
		sprintf(line,"NOWINNER: \"%d\"",jackpot/2);
		gl_log(line);
	}

	free_list();
	#ifdef AUTORESTART
		strcpy(username, A_USER);
		strcpy(group, A_GROUP);	
		startgame(stream);
	#endif
}

/*************************************************************
* Function: is_started()
* Returns : ture if a dicedb exists which means a game started
*           false otherwise.
*************************************************************/

int is_started() {
	struct stat st;
	stat(DICEDB,&st);
	return S_ISREG(st.st_mode);
}


/*************************************************************
* Function: load_users()
* Desc    : Loads data from DICEDB to the linked user list
* Note    : dont forget to free_list() once done using the list.
**************************************************************/

int load_users() {
	FILE *dicedb, *dicejp;
	char cg_creds[LEN], gam1[7], gam2[7];
	gambler *user;

	dicejp=fopen(DICEJP,"r");
	if (dicejp!=NULL) {
		fgets(line,LEN-1,dicejp);
		jackpot=atoi(line);
		fclose(dicejp);
		/* printf("Loaded previous games accumulating jackpot of %dMb !\n",jackpot); */
	}
	dicedb=fopen(DICEDB,"r");
	if (dicedb==NULL) {
		return -1;
	}
        while (fgets(line,LEN-1,dicedb)!=NULL) {
                line[strlen(line)-1]='\0';
                user=new_user();        /* Allocate memory for new user */

                strsplit_s(user->name, line, ' ', 0); /* Get gambler's username */
				strsplit_s(user->group, line, ' ', 1); /* Get gambler's group */
				strsplit_s(cg_creds, line, ' ',2);
                strsplit_s(gam1, line, ' ', 3);
                strsplit_s(gam2, line, ' ', 4);
	
                user->g1 = atoi(gam1);  /* Get gamble dice I  */
                user->g2 = atoi(gam2);  /* Get gamble dice II */
                user->g_creds = atoi(cg_creds);  /* Get amount of credits gambled in bytes*/
                user->credits = get_credits(user->name);
                jackpot += user->g_creds;
                if (user->g1 == dn || user->g2 == dn) {
                        winners++;
						win_creds += user->g_creds;
                        user->winner=1;
                } else  user->winner=0;
        }
        fclose(dicedb);
        /* Finished collecting data */
	return 0;
}


/************************************************************
* Function: get_cretids()
* Gets    : username string
* Returns : credits in bytes of the user, 0 if cant find user
************************************************************/

unsigned long get_credits(const char *username) {
	FILE *fuser;
	char *ptr;

	sprintf(path,"%s/%s",USERPATH,username);
	fuser=fopen(path,"r");
	if (fuser==NULL) {
		printf("Could not open %s's userfile (%s)\n",username,path);
		return 0;
	} 
	while (fgets(line,LEN-1,fuser)!=NULL) {
		if (strncmp(line,"CREDITS",7)==0) {
			line[strlen(line)-1]='\0'; 	/* remove \n */
			if ( (ptr=strchr(line+8,' ')) !=NULL ) ptr = '\0';
				/* ^^ Use the first statsection only */
			return (unsigned long)atoll(line+8);	/* remove "CREDITS " */
		}
	}
	fclose(fuser);
	return 0;
}


/************************************************************
* Function: change_credits()
* Gets    : username string, change_mb integer
* Desc    : changes user's credits by change_mb which can be
*           negative value for deceaseing...
************************************************************/

void change_credits(char *username, float change_mb) {
        FILE *fuser, *tmpfile;
        char path2[MAXPATHLEN], *ptr;
	long long write_creds;
	int fd;

        sprintf(path,"%s/%s",USERPATH,username);
        sprintf(path2,"%s.XXXXXX",path);

        fd = mkstemp(path2);    /* Create temp file to write to */
        if( fd == -1 ) {
                printf("Cant open '%s' for writing!\n",path2);
                return;
        }
        tmpfile = fdopen(fd,"w");
        if (tmpfile==NULL) {
                printf("Error %d: %s (%s tempfile)\n",errno,strerror(errno),path2);
                return;
        }

        fuser=fopen(path,"r");
        if (fuser==NULL) {                                                       
                printf("Could not open %s's userfile (%s)\n",username,path);
                return;
        }
        while (fgets(line,LEN-1,fuser)!=NULL) {
                if (strncmp(line,"CREDITS",7)==0) {
			line[strlen(line)-1]='\0';      /* remove \n */
			ptr = strchr(line+8, ' ');	/* find the end of the first stat section */
			if (ptr != NULL) {
				*ptr = '\0'; 	/* Change the space to null */
				ptr++;		/* Then, Advance the pointer by 1 */
			}
			write_creds = (long long)(atoll(line+8)+(1024*change_mb));
			if (write_creds < 0) write_creds = 0;
			if (ptr == NULL)
				fprintf(tmpfile,"CREDITS %lld\n",write_creds);
			else	fprintf(tmpfile,"CREDITS %lld %s\n",write_creds,ptr);
                } else fputs(line,tmpfile);
        }
        fclose(fuser);
	fclose(tmpfile);
	rename(path2,path);
	chmod(path,0666);

}


/****************************************************************
* Function: strsplit_s()
* Get     : output - initlized string with the same length as str
*           str - the string from which the part will be copied
*           separator - the char which separates the parts
*           index - the # of part to be copied
* Desc    : the function will copy to output, the Nth (index'th) 
*           part of the str using separator as a seprator.
*****************************************************************/

void strsplit_s(char *output, const char *str, const char separator, int index) {
        int idx=0,t=0,start=0,end=0,foundsep=0;

        if (index<0) {
		strcpy(output,str);
		return;
	}
        if (str==NULL) {
		*output='\0';
		return;
	}
	for (t=0 ; *(str+t) != '\0' ; t++) {
                if (*(str+t)==separator) { /* found index */
                        foundsep=1;
                        if ((idx+1)==index) { start=t+1; }
                        else if (idx==index) { end=t-1; break; }
                        idx++;
                }
        }

        if ((foundsep==0) || (index>idx)) { 
		*output='\0';
		return;
	}
        if ((start>0) && (end==0)) end=t;
	strncpy(output,str+start,end-start+1);
	*(output+end-start+1)='\0';

}

void gl_log(char *logline) {
	FILE *log;
	char *sdate;
	time_t k;
	
	k=time(0);
	sdate=ctime(&k);
	*(sdate+24)='\0'; /* remove the \n */
	
	log=fopen(GLFTPDLOG,"a");
	if (log==NULL) {
		printf("Error %d: %s (%s glftpdlog)\n",errno,strerror(errno),GLFTPDLOG);
		return;
	}
	fprintf(log,"%s %s\n",sdate,logline);
	fclose(log);
}

/********************************************************
* Linked list functions.....
* new_user() -> creates a new user, allocates memory, 
*              and return the structure 
* free_list() -> frees the memory taken by the user list
*********************************************************/

gambler *new_user() {
	gambler *new_g, *g_ptr;

	new_g=(gambler *)malloc(sizeof(struct gambler));
	new_g->next=NULL;
	total_gamblers++;

	if (anchor==NULL) anchor=new_g;
	else {
		for(g_ptr=anchor; g_ptr->next != NULL; g_ptr=g_ptr->next);
		g_ptr->next=new_g;
	}
	return new_g;
}

void free_list() {
	gambler *next;
	for(; anchor; anchor=next) {
		next=anchor->next;
		free(anchor);
	}
	/* Reset default settings */
	anchor=NULL;
	total_gamblers=0; winners=0; jackpot=0; win_creds=0;
}
